import math
import pygame
import config
from Displayable import Displayable

class SpriteSequence:
	def __init__(self, fps, sequence, refx, refy, hitboxes = [], loop = True, freeze_frame = None):
		self.fps = fps
		self.sequence = sequence
		self.refx = refx
		self.refy = refy
		self.loop = loop
		self.freeze_frame = freeze_frame

		self.hitboxes = []
		for hb in hitboxes:
			self.hitboxes.append(pygame.rect.Rect(*hb))

class SpriteBase:
	def __init__(self, surface, width, height, frames):
		self.surface = surface
		self.width = width
		self.height = height
		self.frames = frames
		self.sequences = {}
	
	def add_anim_sequence(self, name, seq):
		self.sequences[name] = seq

class Sprite(Displayable):
	def __init__(self, name, seq_name, fps = None):
		Displayable.__init__(self)

		self.base = config.sprite_manager[name]
		self.set_sequence(seq_name, fps)

		self.anim_end_cbfn = None

	def move(self, x, y):
		Displayable.move(self, x, y)
		self.rect.left = x - self.act_seq.refx
		self.rect.top = y - self.act_seq.refy
	
	def move_rel(self, x, y):
		Displayable.move_rel(self, x, y)
		self.rect.move_ip(x, y)

	def set_sequence(self, seq_name, fps = None, start_frame = 0):
		self.act_seq = self.base.sequences[seq_name]

		self.act_frames = map(self.base.surface.subsurface, self.base.frames)

		self.act_frame = start_frame
		if fps is None:
			self.fps = self.act_seq.fps
		else:
			self.fps = fps

		self.start_frame = start_frame
		self.sum_time = 0
		self.freezed = False
		
		self.rect = pygame.rect.Rect(self.x - self.act_seq.refx, self.y - self.act_seq.refy, self.base.width, self.base.height)

	def set_alpha(self, alpha):
		for s in self.act_frames:
			s.set_alpha(alpha)
		
	def update(self, elapsed_msecs):
		self.sum_time += elapsed_msecs
		
		if self.freezed:
			return		
		
		seq_len = len(self.act_seq.sequence)
		next_frame = int(self.start_frame + self.fps * self.sum_time / 1000.0)
		if next_frame >= seq_len:
			if self.act_seq.loop:
				self.act_frame = next_frame % seq_len
				#self.start_frame = self.act_frame
				#self.sum_time -= seq_len * 1000.0 / self.fps
			else:
				self.act_frame = self.act_seq.freeze_frame
				self.freezed = True
				
				if self.anim_end_cbfn is not None:
					self.anim_end_cbfn(*self.anim_end_cbpars)
		else:
			self.act_frame = next_frame
		
	def draw(self, surf):
		if not self.visible:
			return

		if not self.freezed:
			src_surf = self.act_frames[ self.act_seq.sequence[self.act_frame] ]
		else:
			src_surf = self.act_frames[ self.act_seq.freeze_frame ]

		surf.blit(src_surf, (int(self.x - self.act_seq.refx), int(self.y - self.act_seq.refy)))

		#pygame.draw.rect(surf, (255, 0, 0, 255), self.rect, 1)
		
		#for r in self.act_seq.hitboxes:
		#	self_moved_hb = r.move(self.x - self.act_seq.refx, self.y - self.act_seq.refy)
		#	pygame.draw.rect(surf, (0, 255, 0, 255), self_moved_hb, 1)

	def collides(self, splist):
		collide_list = []
		
		for s in splist:
			for self_hb in self.act_seq.hitboxes:
				self_moved_hb = self_hb.move(self.x - self.act_seq.refx, self.y - self.act_seq.refy)
				for other_hb in s.act_seq.hitboxes:
					other_moved_hb = other_hb.move(s.x - s.act_seq.refx, s.y - s.act_seq.refy)

					if self_moved_hb.colliderect(other_moved_hb):
						collide_list.append( (s, self_moved_hb.clip(other_moved_hb)) )

		return collide_list
	
	def collide_pt(self, pos):
		for self_hb in self.act_seq.hitboxes:
			self_moved_hb = self_hb.move(self.x - self.act_seq.refx, self.y - self.act_seq.refy)

			if self_moved_hb.collidepoint(pos[0], pos[1]):
				return True

		return False

	def screen_position(self):
		scr_rect = pygame.rect.Rect(0, 0, config.SCREEN_WIDTH, config.SCREEN_HEIGHT)
		if scr_rect.contains(self.rect):
			return 1

		if scr_rect.clip(self.rect).size == (0, 0):
			return -1

		return 0

	def anim_end_callback(self, cbfn, *cb_pars):
		self.anim_end_cbfn = cbfn
		self.anim_end_cbpars = cb_pars

class MovingSprite(Sprite):
	def __init__(self, name, seq_name, fps = None):
		Sprite.__init__(self, name, seq_name, fps)

		self.out_of_screen_cbfn = None
		
		self.dont_move = False

	def set_destination(self, dest_x, dest_y, pixels_per_sec):
		self.dest_x = dest_x
		self.dest_y = dest_y
		self.pps = pixels_per_sec
		
	def stop(self):
		self.dont_move = True

	def start(self):
		self.start_x = self.x
		self.start_y = self.y
		self.l = math.sqrt((self.start_x - self.dest_x)**2 + (self.start_y - self.dest_y)**2)
		
		self.move_sum_time = 0
		self.dont_move = False

	def update(self, elapsed_msecs):
		Sprite.update(self, elapsed_msecs)
		if self.dont_move:
			return

		self.move_sum_time += elapsed_msecs
		
		try:
			factor = self.pps * self.move_sum_time / 1000.0 / self.l
		except ZeroDivisionError:
			return
			
		#if factor < 1.0:
		self.move(self.start_x + (self.dest_x - self.start_x) * factor, self.start_y + (self.dest_y - self.start_y) * factor)
		#else:
		#	self.move(self.dest_x, self.dest_y)
		#	self.stop()

		if self.screen_position() == -1:
			if self.out_of_screen_cbfn is not None:
				self.out_of_screen_cbfn(*self.out_of_screen_cbpars)
				
			self.stop()

	def out_of_screen_callback(self, cbfn, *pars):
		self.out_of_screen_cbfn = cbfn
		self.out_of_screen_cbpars = pars

class FuzzySprite(Sprite):
	
	def __init__(self, name, seq_name, fps = None):
		Sprite.__init__(self, name, seq_name, fps)
		self.amplitude = (10, 15)
		self.wfreq = (1.0, 0.5)
		self.phases = (0.4, 0.1)
		self.lastDelta = [0,0]
	
	def setFuzzy(amplitude, wfreq, phases):
		self.amplitude = amplitude
		self.wfreq = wfreq
		self.phase = phases
			
	def update(self, elapsed_msecs):
		Sprite.update(self, elapsed_msecs)
		self.x -= self.lastDelta[0]
		self.y -= self.lastDelta[1]
		
		for i in (0,1):
			self.lastDelta[i] = self.amplitude[i] * math.sin(self.wfreq[i]*self.sum_time/1000.0 + self.phases[i])
		
		
		self.x += self.lastDelta[0]
		self.y += self.lastDelta[1]

class AccelSprite(Sprite):
	def __init__(self, name, seq_name, top_speed, fps = None):
		Sprite.__init__(self, name, seq_name, fps)
		
		self.out_of_screen_cbfn = None

		self.top_speed = top_speed

		self.set_speed(0.0, 0.0)
		self.set_accel(0.0, 0.0)

	def set_accel(self, accel_x, accel_y):
		self.accel_x = accel_x
		self.accel_y = accel_y
	
	def add_accel(self, accel_x, accel_y):
		self.accel_x += accel_x
		self.accel_y += accel_y
	
	def set_speed(self, speed_x, speed_y):
		self.speed_x = speed_x
		self.speed_y = speed_y
	
	def add_speed(self, speed_x, speed_y):
		self.speed_x += speed_x
		self.speed_y += speed_y

	def set_destination(self, dest_x, dest_y, pixels_per_sec):
		self.dest_x = dest_x
		self.dest_y = dest_y
		self.pps = pixels_per_sec

	def update(self, elapsed_msecs):
		Sprite.update(self, elapsed_msecs)

		self.speed_x += self.accel_x * elapsed_msecs / 1000.0
		self.speed_y += self.accel_y * elapsed_msecs / 1000.0

		speed = math.sqrt(self.speed_x**2 + self.speed_y**2)
		if speed > self.top_speed:
			self.speed_x *= self.top_speed / speed
			self.speed_y *= self.top_speed / speed
			speed = self.top_speed
			
		self.act_speed = speed

		#self.move_rel(self.speed_x * elapsed_msecs / 1000.0, self.speed_y * elapsed_msecs / 1000.0)
		self.move(self.x + self.speed_x * elapsed_msecs / 1000.0, self.y + self.speed_y * elapsed_msecs / 1000.0)

		if self.screen_position() == -1:
			if self.out_of_screen_cbfn is not None:
				self.out_of_screen_cbfn(*self.out_of_screen_cbpars)
				
	def out_of_screen_callback(self, cbfn, *pars):
		self.out_of_screen_cbfn = cbfn
		self.out_of_screen_cbpars = pars

class MeterBar(Displayable):
	def __init__(self, width, height, max, vertical = False, inCol=(0, 255, 0), bgCol=(255,0,0), outCol = (255,255,255)):
		Displayable.__init__(self)
		
		self.width = width
		self.height = height
		self.max = max
		self.act = 0
		self.vertical = vertical
		
		self.inCol = inCol
		self.bgCol = bgCol
		self.outCol = outCol
		
		self.rect = pygame.rect.Rect(self.x, self.y, self.width, self.height)

	def move(self, new_x, new_y):
		Displayable.move(self, new_x, new_y)
		self.rect.left = new_x
		self.rect.top = new_y
	
	def move_rel(self, x, y):
		Displayable.move_rel(self, x, y)
		self.rect.move_ip(x, y)

	def set_amount(self, act):
		self.act = act

	def draw(self, surf):
		pygame.draw.rect(surf, self.outCol, self.rect, 1)
		
		r = self.rect.inflate(-2, -2)
		pygame.draw.rect(surf, self.bgCol, r, 0)

		if self.vertical:
			oldh = r.height
			r.height = int(r.height * float(self.act) / self.max)
			r.y += oldh - r.height
		else:
			r.width = int(r.width * float(self.act) / self.max)
		pygame.draw.rect(surf, self.inCol, r, 0)

class TextSprite(Displayable):
	def __init__(self, x, y, text, color, size, font, antialias = True, shadow = None, shadow_offs = 2):
		Displayable.__init__(self)

		self.text = None
		self.color = None
		self.size = None
		self.font_name = None
		self.font = None
		self.surf = None
		self.shadow = shadow
		self.shadow_offs = shadow_offs
		self.anitalias = antialias

		self.set_text(text, color, size, font, antialias, shadow, shadow_offs)
		self.set_alignment(-1, -1)
		self.move(x, y)

	def set_text(self, text, color = None, size = None, font = None, antialias = None, shadow = None, shadow_offs = None):
		self.text = text
		
		if color is not None:
			self.color = color
		
		if size is not None:
			self.size = size
		
		if font is not None:
			self.font_name = font
		
		if shadow is not None:
			self.shadow = shadow
		
		if shadow_offs is not None:
			self.shadow_offs = shadow_offs

		if antialias is not None:
			self.antialias = antialias

		self.font = config.font_manager[self.font_name][self.size]
		self.surf = self.font.render(self.text, self.antialias, self.color)

		if self.shadow is not None:
			w = self.surf.get_width()
			h = self.surf.get_height()
			
			sh_surf = pygame.surface.Surface((w + self.shadow_offs, h + self.shadow_offs), 0)
			sh_surf.fill((11, 55, 99))
			sh_surf.set_colorkey((11, 55, 99))
			
			shadow = self.font.render(self.text, self.antialias, self.shadow)
			sh_surf.blit(shadow, (self.shadow_offs, self.shadow_offs))
			sh_surf.blit(self.surf, (0, 0))
			self.surf = sh_surf
			
		self.width = self.surf.get_width()
		self.height = self.surf.get_height()

	def set_alignment(self, halign, valign):
		if halign is not None:
			self.halign = halign
		
		if valign is not None:
			self.valign = valign

		if self.halign == -1:
			self.xoffs = 0
		elif self.halign == 0:
			self.xoffs = - self.width / 2
		else:
			self.xoffs = - self.width
		
		if self.valign == -1:
			self.yoffs = 0
		elif self.valign == 0:
			self.yoffs = - self.height / 2
		else:
			self.yoffs = - self.height	
		
	def draw(self, surf):
		surf.blit(self.surf, (self.x + self.xoffs, self.y + self.yoffs))

class HintTextSprite(TextSprite):
	def __init__(self, x, y, text, color, size, font, fade_in_time, display_time, fade_out_time):
		#TextSprite.__init__(self, x, y, text, color, size, font, False, (0, 0, 0), 2)
		TextSprite.__init__(self, x, y, text, color, size, font, False, (0, 0, 0), 2)
		
		self.set_alignment(0, 0)

		self.fade_in_time = fade_in_time
		self.display_time = display_time
		self.fade_out_time = fade_out_time

		self.alpha = 0
		self.sum_time = 0
		self.phase = 0

		self.cbfn = None

	def disappear_callback(self, cbfn, *pars):
		self.cbfn = cbfn
		self.cbpars = pars

	def update(self, elapsed_msecs):
		if self.phase is None:
			return
			
		self.sum_time += elapsed_msecs
		if self.phase == 0:
			self.alpha = int(255.0 * self.sum_time / (self.fade_in_time * 1000.0))
			if self.alpha >= 255:
				self.alpha = 255
				self.sum_time = 0
				self.phase = 1

		elif self.phase == 1:
			if self.sum_time >= self.display_time * 1000.0:
				self.sum_time = 0
				self.phase = 2

		if self.phase == 2:
			self.alpha = 255 - int(255.0 * self.sum_time / (self.fade_out_time * 1000.0))
			if self.alpha <= 0:
				self.alpha = 0
				self.sum_time = 0
				self.phase = None
				
				if self.cbfn is not None:
					self.cbfn(*self.cbpars)
		
		TextSprite.update(self, elapsed_msecs)
	
	def draw(self, surf):
		self.surf.set_alpha(self.alpha)
		surf.blit(self.surf, (self.x + self.xoffs, self.y + self.yoffs))

class BackgroundSprite(Displayable):
	def __init__(self, top_color, bottom_color):
		disp_surf = pygame.display.get_surface()
		width, height = disp_surf.get_size()
		
		self.surf = pygame.surface.Surface((width, height), 0, disp_surf)

		act_color = [0, 0, 0]
		
		for y in xrange(height):
			for c in xrange(3):
				act_color[c] = top_color[c] + (bottom_color[c] - top_color[c]) * y / height

			pygame.draw.line(self.surf, act_color, (0, y), (width - 1, y), 1)

	def draw(self, surf):
		surf.blit(self.surf, (0, 0))

